import logging
import os
import sys
from argparse import ArgumentParser
from pathlib import Path

from peak import JID
from peak import __name__ as peak_name
from peak import __version__ as version
from peak import configure_cli_logger
from peak.cli import df, run, send, start

_logger = logging.getLogger(__name__)


def main(args=None):
    try:
        configure_cli_logger()
        _logger.info("initiating PEAK")
        _main(args)
        _logger.info("PEAK terminated")
    except Exception as e:
        _logger.critical(e, exc_info=1)
    except KeyboardInterrupt:
        _logger.info("PEAK terminated (KeyboardInterrupt)")


def _main(args=None):
    _logger.info("parsing console arguments")
    parser = ArgumentParser(prog=peak_name)
    parser.add_argument("--version", action="version", version=version)
    subparsers = parser.add_subparsers(required=True)

    # parser for "df" command
    df_parser = subparsers.add_parser(
        name="df", help="execute Directory Facilitator agent"
    )
    df_parser.add_argument(
        "-domain",
        type=str,
        default="localhost",
        help="XMPP server domain (default: localhost)",
    )
    df_parser.add_argument(
        "--verify_security", action="store_true", help="verify SLL certificates"
    )
    df_parser.add_argument(
        "-log_level",
        type=str.upper,
        default="INFO",
        help="PEAK logging level (default: INFO)",
    )
    df_parser.add_argument(
        "-port",
        type=str,
        default="10000",
        help="REST API port (default: 10000)",
    )
    df_parser.set_defaults(func=df.exec)

    # parser for the "run" command
    run_parser = subparsers.add_parser(
        name="run",
        help="execute a single-agent system using a Python script",
    )
    run_parser.add_argument(
        "file",
        type=Path,
        help="Python file containing the class of the agent to be executed (the same name must be used in the class and in the file) ",
    )
    run_parser.add_argument(
        "-jid", type=JID.fromstr, help="agent XMPP ID", required=True
    )
    run_parser.add_argument(
        "-clones",
        type=int,
        default=1,
        help="number of clones",
    )
    run_parser.add_argument(
        "-log_level",
        type=str.upper,
        default="INFO",
        help="PEAK logging level (default: INFO)",
    )
    run_parser.add_argument(
        "-f",
        "--log_folder",
        type=Path,
        default=Path(os.getcwd()).joinpath("logs"),
        help="logs folder (default: folder named 'logs' in current working directory)",
    )
    run_parser.add_argument(
        "-a",
        "--log_file_mode",
        type=str,
        default="a",
        help="mode for logs' file (default: append)",
    )
    run_parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=5222,
        help="XMPP server port (default: 5222)",
    )
    run_parser.add_argument(
        "--debug_mode",
        action="store_true",
        help="mode for logs' file (default: false)",
    )
    run_parser.add_argument(
        "--verify_security", action="store_true", help="verify SLL certificates"
    )
    run_parser.set_defaults(func=run.execute_agent)

    # parser for the "start" command
    start_parser = subparsers.add_parser(
        name="start",
        help="execute agents using an YAML configuration file",
    )
    start_parser.add_argument(
        "file",
        type=Path,
        help="YAML configuration file",
    )
    start_parser.set_defaults(func=start.execute_config_file)

    # parser for the "send" command
    send_parser = subparsers.add_parser(
        name="send",
        help="sends a message to an agent",
    )
    send_parser.add_argument(
        "-to", type=JID.fromstr, help="Jabber ID of the receiver agent", required=True
    )
    send_parser.add_argument(
        "-sender", type=JID.fromstr, help="Jabber ID of the sender agent (optional)"
    )
    send_parser.add_argument("-body", type=str, help="message body (optional)")
    send_parser.add_argument(
        "--thread", type=str, help="thread ID for the message (optional)"
    )
    send_parser.add_argument(
        "-metadata",
        type=dict,
        help="message's metadata (optional)",
    )
    send_parser.set_defaults(func=send.send_message)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(0)

    args = parser.parse_args(args)
    args.func(**vars(args))



from typing import Type

from spade.message import MessageBase as _MessageBase
from spade.template import Template as SpadeTemplate

from peak.logging import getLogger

logger = getLogger(__name__)


class Template(SpadeTemplate):
    """PEAK's template for matching messages."""

    def match(self, message: Type["_MessageBase"]) -> bool:
        """
        Returns wether a message matches with this message or not.
        The message can be a Message object or a Template object.
        Ignores reource from JID if it is not present.

        Args:
          message (peak.message.Message): the message to match to

        Returns:
          bool: wether the message matches or not

        """
        if self.to:
            if self.to.domain != message.to.domain:
                return False
            if self.to.localpart and message.to.localpart != self.to.localpart:
                return False
            if self.to.resource and message.to.resource != self.to.resource:
                return False

        if self.sender:
            if self.sender.domain != message.sender.domain:
                return False
            if (
                self.sender.localpart
                and message.sender.localpart != self.sender.localpart
            ):
                return False
            if self.sender.resource and message.sender.resource != self.sender.resource:
                return False

        if self.body and message.body != self.body:
            return False

        if self.thread and message.thread != self.thread:
            return False

        for key, value in self.metadata.items():
            if message.get_metadata(key) != value:
                return False

        logger.debug(f"message matched {self} == {message}")
        return True



import json

from peak import CyclicBehaviour, Template, getLogger

logger = getLogger(__name__)


class SearchCommunity(CyclicBehaviour):
    """Handles all the requests to search for groups."""

    async def on_start(self) -> None:
        logger.debug(f"starting {self.__class__.__name__}")
        template = Template()
        template.set_metadata("resource", "searchgroup")
        self.set_template(template)

    async def run(self) -> None:
        msg = await self.receive(60)
        if msg and (meta_tags := msg.get_metadata("tags")):
            tags = json.loads(meta_tags)
            communities: set = self.agent.ecosystemhierarchy_data["tags"][tags[0]]
            for tag in tags[1:]:
                communities = communities.intersection(
                    self.agent.ecosystemhierarchy_data["tags"][tag]
                )
            res = msg.make_reply()
            res.set_metadata("communities", json.dumps(list(communities)))
            await self.send(res)



from argparse import ArgumentTypeError
from pathlib import Path

from peak import JID
from peak.bootloader import bootloader


def execute_agent(
    file: Path,
    jid: JID,
    clones: int,
    log_level: str,
    log_folder: Path,
    log_file_mode: str,
    debug_mode: bool = False,
    port: int = 5222,
    verify_security: bool = False,
    *args,
    **kargs,
):
    """Executes a single agent.

    Args:
        file: Path to the agent's python file.
        jid: JID of the agent.
        clones: Number of clones to be made.
        log_level: Logging level.
        verify_security: Verifies the SSL certificates.
    """
    # TODO: verify at argsparser level
    if file and not file.is_file():
        raise ArgumentTypeError(f"file argument must be a python file, not '{file}'")

    kwargs = {
        "file": file,
        "jid": jid,
        "cid": 0,
        "log_level": log_level,
        "log_folder": log_folder,
        "log_file_mode": log_file_mode,
        "debug_mode": debug_mode,
        "port": port,
        "verify_security": verify_security,
    }

    name = jid.localpart
    agents = []
    for cid in range(clones):
        agent = kwargs.copy()
        agents.append(agent)
        kwargs["jid"] = kwargs["jid"].replace(localpart=f"{name}{cid}")
        kwargs["cid"] = cid
    bootloader(agents)



import json

from peak import DF, Message
from peak.core import OneShotBehaviour


class CreateGraph(OneShotBehaviour):
    """Creates a graph in the dashboard.

    Sends the graph configuration and data to the DF so he can host it.
    """

    def __init__(self, id: str, graph: dict):
        super().__init__()
        self.id = id
        self.graph = graph

    async def run(self) -> None:
        msg = Message()
        msg.to = DF.name(self.agent.jid.domain)
        msg.body = f"Create graph {self.id}"
        msg.metadata = {
            "resource": "graph",
            "action": "create",
            "id": self.id,
            "graph": json.dumps(self.graph),
        }
        await self.send(msg)



import asyncio as _asyncio

from peak import Message, PeriodicBehaviour


class PeriodicClock(PeriodicBehaviour):
    """Handles the clock of the simulation.

    This clock tracks the number of the current period
    throughout the simulation.
    """

    def __init__(
        self,
        group_jid: str,
        n_agents: int,
        periods: int,
        time_per_period: float,
        start_at=None,
    ):
        super().__init__(time_per_period, start_at=start_at)
        self.group_jid = group_jid
        self.n_agents = n_agents
        self.periods = periods

    async def on_start(self):
        self.logger.info("Waiting for all agents to enter the group...")
        while not len(await self.agent.group_members(self.group_jid)) >= self.n_agents:
            await _asyncio.sleep(1)
        self.current_period = 0
        self.logger.info("Starting simulation...")

    async def run(self):
        msg = Message()
        msg.to = self.group_jid
        if self.current_period >= self.periods:
            msg.set_metadata("sync", "stop")
            self.kill()
        else:
            self.logger.info(f"Period {self.current_period}")
            msg.body = f"Period {self.current_period}"
            msg.set_metadata("sync", "step")
            msg.set_metadata("period", str(self.current_period))
        await self.send_to_group(msg)
        self.current_period += 1

    async def on_end(self):
        self.logger.info("Ending simulation...")
        await self.agent.stop()



import logging
import sys
from os import PathLike
from typing import Union

FORMATTER = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")


def configure_cli_logger() -> logging.Logger:
    """Configure logger for command line interface."""
    logger = logging.getLogger("peak")
    logger.setLevel(logging.INFO)
    logger.handlers.clear()
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    logger.addHandler(console_handler)
    return logger


def configure_single_agent_logging(
    log_level: Union[int, str], filename: Union[str, PathLike], mode: str
) -> logging.Logger:
    """Configure logging for single agent."""
    logger = logging.getLogger("peak")
    logger.setLevel(log_level)
    logger.propagate = False
    logger.handlers.clear()
    file_handler = logging.FileHandler(filename, mode)
    file_handler.setFormatter(FORMATTER)
    logger.addHandler(file_handler)

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    logger.addHandler(console_handler)


def configure_multiple_agent_logging(
    log_level: Union[int, str], filename: Union[str, PathLike], mode: str
) -> logging.Logger:
    logger = logging.getLogger("peak")
    logger.setLevel(log_level)
    logger.propagate = False
    logger.handlers.clear()
    file_handler = logging.FileHandler(filename, mode)
    file_handler.setFormatter(FORMATTER)
    logger.addHandler(file_handler)

    logger = logging.getLogger("peak.main")
    logger.setLevel(log_level)
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    logger.addHandler(console_handler)


def configure_debug_mode(
    log_level: Union[int, str], filename: Union[str, PathLike], mode: str
):
    """Configure 'root' logger to consume logs from all other modules
    using the 'peak' logger file handler. Only logs them to the file."""
    logger = logging.getLogger()
    logger.setLevel(log_level)
    file_handler = logging.FileHandler(filename, mode)
    file_handler.setFormatter(FORMATTER)
    logger.addHandler(file_handler)


def getLogger(name: str = None) -> logging.Logger:
    """Same as getFileLogger, just renamed it."""
    return getFileLogger(name)


def getFileLogger(name: str = None) -> logging.Logger:
    """Get 'peak' logger or a sublogger of 'peak.{name}'.
    Prints to the file. It only prints in the terminal if there
    is one agent running."""
    if name is None:
        return logging.getLogger("peak")
    return logging.getLogger(f"peak.{name}")


def getMainLogger(name: str = None) -> logging.Logger:
    """Get 'peak.main' logger. Always prints in the terminal."""
    if name is None:
        return logging.getLogger("peak.main")
    return logging.getLogger(f"peak.main.{name}")


def log(msg: str, level: Union[int, str] = logging.INFO):
    """Logs to the file. Default level info."""
    logging.getLogger("peak").log(level, msg)


def debug(msg: str, level: Union[int, str] = logging.DEBUG):
    """Logs to the terminal. Default level debug."""
    logging.getLogger("peak.main").log(level, msg)



from datetime import datetime, timedelta

from peak import Agent
from peak.agents.synchronizer.behaviors import DateTimeClock, PeriodicClock


class Synchronizer(Agent):
    """Synchronizes the syncagents.

    The Synchronizer creates a group of agents, awaits for
    the agents to join the group and starts the clock of the
    simulation.
    """

    async def sync_group_period(
        self, group_jid: str, n_agents: int, interval: float, periods: int
    ):
        """Synchronizes a group of agents.

        The clock is based on the number of the current period.

        Args:
            group_jid: Identifier of the XMPP group to be synchronized.
            n_agents: Number of agents to be synchronized. Synchronizer
                awaits for this number of agents to join the group before
                it starts the simulation.
            interval: Time in seconds between each period.
            periods: Number of periods to simulate.
        """
        await self.join_group(group_jid)
        self.add_behaviour(PeriodicClock(group_jid, n_agents, periods, interval))

    async def sync_group_time(
        self,
        group_jid: str,
        n_agents: int,
        initial_datetime: datetime,
        end_datetime: datetime,
        internal_interval: timedelta,
        external_period_time: float,
        start_at: datetime = None,
    ):
        """Synchronizes a group of agents.

        Here two time dimensions are created. One is the real-time at which
        the clock of the Synchronizer will run. The other is the fictional
        datetime created inside the simulation. For example, one second
        can correspond to one day inside the simulation.

        Args:
            group_jid: Identifier of the XMPP group to be synchronized.
            n_agents: Number of agents to be synchronized. Synchronizer
                awaits for this number of agents to join the group before
                it starts the simulation.
            initial_datetime: Defines the initial date and time inside the
                simulation.
            end_datetime: Defines the date and time at which the simulation ends.
            internal_interval: Time between each period relative to the initial and
                end datetimes.
            interval: Time in seconds between each period relative to the Synchronizers
                clock.
            start_at: Schedules the simulation to start at a given time. If None the
                simulation starts right away.
        """
        await self.join_group(group_jid)
        self.add_behaviour(
            DateTimeClock(
                group_jid,
                n_agents,
                initial_datetime,
                end_datetime,
                internal_interval,
                external_period_time,
                start_at,
            )
        )



import json

from peak import DF, Message
from peak.core import OneShotBehaviour


class JoinCommunity(OneShotBehaviour):
    """Joins a community using a JID."""

    def __init__(self, path: str, domain: str, tags: list = []):
        super().__init__()
        self.path = path
        self.domain = domain
        self.tags = tags

    async def run(self):
        msg = Message()
        msg.to = DF.name(self.agent.jid.domain)
        msg.set_metadata("resource", "treehierarchy")
        msg.set_metadata("path", self.path)
        msg.set_metadata("domain", self.domain)
        msg.set_metadata("tags", json.dumps(self.tags))
        await self.send(msg)
        nodes = self.path.split("/")
        for node in nodes[:-1]:
            await self.join_community(node + "_down@" + self.domain)
        await self.join_community(nodes[-1] + "_down@" + self.domain)
        await self.join_community(nodes[-1] + "@" + self.domain)



import aiohttp_cors
from aioxmpp import JID

from peak import Agent

from .behaviors import CreateGraph, EcosystemHierarchy, SearchCommunity


class DF(Agent):
    """Directory Facilitator.

    This agent makes available the data that it gathers from multi-agent systems
    publicly through a REST API. It also provides a Yellow Page Service to the
    agents.
    """

    def __init__(self, domain, verify_security, port):
        super().__init__(
            JID.fromstr("df@" + domain + "/admin"), verify_security=verify_security
        )
        self.port = port

    @classmethod
    def name(cls, domain: str) -> str:
        """Retrieves the JID of the DF based on the agent domain.

        Args:
            domain: server's name

        Returns:
            The string of the complete JID of the DF.
        """
        return "df@" + domain + "/admin"

    async def setup(self):
        self.ecosystemhierarchy_data = {
            "nodes": set(),
            "links": set(),
            "categories": set(),
            "node_members": {},
            "tags": {},
        }
        self.dataanalysis_data = dict()
        self.group_tags = dict()

        self.add_behaviour(EcosystemHierarchy())
        self.add_behaviour(SearchCommunity())
        self.add_behaviour(CreateGraph())

        # Create routes.
        self.web.add_get("/groups", self.get_groups, template=None)
        self.web.add_get("/groups/refresh", self.refresh_groups, template=None)
        self.web.add_get("/plots", self.get_plots, template=None)

        # Configure default CORS settings.
        cors = aiohttp_cors.setup(
            self.web.app,
            defaults={
                "*": aiohttp_cors.ResourceOptions(
                    allow_credentials=True,
                    expose_headers="*",
                    allow_headers="*",
                )
            },
        )

        # Configure CORS on all routes.
        for route in list(self.web.app.router.routes()):
            cors.add(route)

        # Start web API
        self.web.start("0.0.0.0", port=self.port)
        self.logger.info("REST API running on port " + self.port)

    async def get_groups(self, request):
        return {
            "nodes": list(self.ecosystemhierarchy_data["nodes"]),
            "links": list(self.ecosystemhierarchy_data["links"]),
            "categories": list(self.ecosystemhierarchy_data["categories"]),
            "node_members": self.ecosystemhierarchy_data["node_members"],
        }

    async def refresh_groups(self, request):
        pass

    async def get_plots(self, request):
        return self.dataanalysis_data



import logging
import sys
from pathlib import Path

import yaml

from peak import JID
from peak.bootloader import bootloader

_logger = logging.getLogger(__name__)


def execute_config_file(file: Path, *args, **kargs):
    """Executes agents using a YAML configuration file.

    Args:
        file: Path to the agent's python file.
    """

    _logger.info("parsing YAML configuration file")
    defaults = {
        "file": None,
        "domain": "localhost",
        "resource": "main",
        "ssl": False,
        "log_level": "info",
        "log_folder": "logs",
        "log_file_mode": "a",
        "debug_mode": False,
        "port": 5222,
        "clones": 1,
        "verify_security": False,
    }
    agents = []

    with file.open() as f:
        yml = yaml.full_load(f)
    sys.path.append(
        str(file.parent.absolute())
    )  # imports any python modules in the parent folder of the yaml file

    if "defaults" in yml:
        defaults = defaults | yml["defaults"]
    if "agents" not in yml:
        raise Exception("YAML: 'agents' argument required")
    for agent_name, agent_args in yml["agents"].items():
        if agent_args is not None:
            agent_args = defaults | agent_args
        else:
            agent_args = defaults
        if agent_args["file"] is None:
            raise Exception(f"{agent_name}: file argument required")
        if agent_args["domain"] is None:
            raise Exception(f"{agent_name}: domain argument required")
        kwargs = {
            "file": file.parent.joinpath(agent_args["file"]),
            "jid": JID(agent_name, agent_args["domain"], agent_args["resource"]),
            "cid": 0,
            "log_level": agent_args["log_level"].upper(),
            "log_folder": file.parent.joinpath(agent_args["log_folder"]),
            "log_file_mode": agent_args["log_file_mode"],
            "verify_security": agent_args["ssl"],
            "debug_mode": agent_args["debug_mode"],
            "port": agent_args["port"],
        }
        if agent_args["clones"] > 1:
            for cid in range(agent_args["clones"]):
                kwargs["jid"] = kwargs["jid"].replace(localpart=f"{agent_name}{cid}")
                kwargs["cid"] = cid
                agent = kwargs.copy()
                agents.append(agent)
        else:
            agents.append(kwargs)

    _logger.info("YAML configuration file parsed")
    bootloader(agents)



from peak import Message, OneShotBehaviour, getLogger

logger = getLogger(__name__)


class SendMessage(OneShotBehaviour):
    """Sends a message to another agent."""

    def __init__(self, msg: Message):
        super().__init__()
        self.msg = msg

    async def run(self) -> None:
        await self.send(self.msg)
        logger.info(f"message sent: {self.msg}")
        await self.agent.stop()



from datetime import datetime

from peak import CyclicBehaviour


class StepBehaviour(CyclicBehaviour):
    async def on_start(self):
        self.logger.info("Waiting for simulation to start...")

    async def run(self):
        msg = await self.receive(10)
        if msg:
            if msg.get_metadata("sync") == "step":
                period = int(msg.get_metadata("period"))
                time = None
                if msg.get_metadata("time"):
                    time = datetime.strptime(
                        msg.get_metadata("time"), "%Y-%m-%d %H:%M:%S"
                    )
                await self.agent.step(period, time)
                self.logger.info(msg.body)
            if msg.get_metadata("sync") == "stop":
                self.logger.info("Simulation ended.")
                self.kill()

    async def on_end(self):
        await self.agent.stop()



from time import sleep

from peak import JID, DummyAgent, Message


def send_message(
    to: JID, sender: JID, body: str, thread: str, metadata: dict, *args, **kargs
):
    if sender is None:
        sender = f"dummyagent@{to.domain}/main"
    msg = Message(
        to=str(to), sender=str(sender), body=body, thread=thread, metadata=metadata
    )
    print(to, sender, body, msg.sender, type(msg.sender))
    da = DummyAgent(jid=msg.sender, message=msg)
    da.start().result()
    while da.is_alive():
        sleep(0.5)



from abc import ABCMeta as _ABCMeta
from abc import abstractmethod
from datetime import datetime

from aioxmpp import JID

from peak import Agent, Template

from .behaviors import StepBehaviour


class SyncAgent(Agent, metaclass=_ABCMeta):
    """Is synchronized by the Synchronizer.

    Every agent that needs to be synchronized needs
    to extend this class.
    """

    def __init__(self, jid: JID, verify_security: bool = False):
        """Inits the SyncAgent with the JID provided.

        Args:
            jid (:obj:`JID`): Agent's XMPP identifier.
            verify_security (bool, optional): If True, verifies the SSL certificates.
                Defaults to False.
        """
        super().__init__(jid, verify_security)
        template_step = Template()
        template_step.set_metadata("sync", "step")
        template_stop = Template()
        template_stop.set_metadata("sync", "stop")
        template = template_step | template_stop
        self.add_behaviour(StepBehaviour(), template)

    @abstractmethod
    async def step(self, period: int, time: datetime = None):
        """Executed at each tick of the Synchronizer clock.

        To be implemented by the user.

        Args:
            period (int): Number of the current period.
            time (datetime, optional): Current datetime inside the simulation. It must be
                configured in the Synchronizer."""
        raise NotImplementedError()



from peak import Agent, Message

from .behaviors import SendMessage


class DummyAgent(Agent):
    """Dummy agent for sending messages."""

    def __init__(self, message: Message, *args, **kargs):
        super().__init__(*args, **kargs)
        self.message = message

    async def setup(self):
        self.add_behaviour(SendMessage(self.message))



import json

from peak import CyclicBehaviour, Template, getLogger

logger = getLogger(__name__)


class EcosystemHierarchy(CyclicBehaviour):
    """Manages the group structure of all the multi-agent systems."""

    async def on_start(self):
        template = Template()
        template.set_metadata("resource", "treehierarchy")
        self.set_template(template)

    async def run(self):
        msg = await self.receive(60)
        if msg:
            logger.debug("message received")
            path = msg.get_metadata("path")
            domain = msg.get_metadata("domain")
            if meta_tags := msg.get_metadata("tags"):
                tags = json.loads(meta_tags)
            nodes = path.split("/")
            logger.debug("nodes: " + str(nodes))
            level = "level"

            if (
                msg.get_metadata("leave")
                and msg.sender
                in self.agent.ecosystemhierarchy_data["node_members"][nodes[-1]]
            ):
                logger.debug(str(msg.sender) + " leaving " + path)
                self.agent.ecosystemhierarchy_data["node_members"][nodes[-1]].remove(
                    msg.sender
                )
                nodes = nodes[::-1]
                # remove empty nodes and links
                for i, node in enumerate(nodes):
                    if len(
                        self.agent.ecosystemhierarchy_data["node_members"][node]
                    ) == 0 and not any(
                        node == source
                        for source, _ in self.agent.ecosystemhierarchy_data["links"]
                    ):
                        self.agent.ecosystemhierarchy_data["node_members"].pop(
                            node
                        )  # this line can be removed, there is no need in removing the node from the dict
                        self.agent.ecosystemhierarchy_data["nodes"].remove(
                            (node, level + str(len(nodes) - 1 - i), domain)
                        )
                        if i + 1 < len(nodes):
                            self.agent.ecosystemhierarchy_data["links"].remove(
                                (nodes[i + 1], node)
                            )
                existing_categories = set()

                # remove categories if empty
                for _, level, _ in self.agent.ecosystemhierarchy_data["nodes"]:
                    existing_categories.add(level)
                difference = self.agent.ecosystemhierarchy_data[
                    "categories"
                ].difference(existing_categories)
                if any(difference):
                    self.agent.ecosystemhierarchy_data["categories"] -= difference

            else:
                logger.debug(str(msg.sender) + " entering " + path)
                last = None
                for i, node in enumerate(nodes):
                    self.agent.ecosystemhierarchy_data["nodes"].add(
                        (node, level + str(i), domain)
                    )
                    if node not in self.agent.ecosystemhierarchy_data["node_members"]:
                        self.agent.ecosystemhierarchy_data["node_members"][node] = []
                    self.agent.ecosystemhierarchy_data["categories"].add(level + str(i))
                    if last != None:
                        self.agent.ecosystemhierarchy_data["links"].add(
                            (
                                last,
                                node,
                            )
                        )
                    last = node
                self.agent.ecosystemhierarchy_data["node_members"][last].append(
                    msg.sender
                )
                for tag in tags:
                    if tag not in self.agent.ecosystemhierarchy_data["tags"]:
                        self.agent.ecosystemhierarchy_data["tags"][tag] = set()
                    self.agent.ecosystemhierarchy_data["tags"][tag].add(last)



import asyncio as _asyncio
from datetime import datetime, timedelta

from peak import Message, PeriodicBehaviour


class DateTimeClock(PeriodicBehaviour):
    """Handles the clock of the simulation.

    This clock tracks the current date and time of the
    simulation throughout its execution.
    """

    def __init__(
        self,
        jid,
        n_agents: int,
        initial_time: datetime,
        end_time: datetime,
        period_time_simulated: timedelta,
        period_time_real: float,
        start_at: datetime = None,
    ):
        super().__init__(period_time_real, start_at=start_at)
        self.group_jid = jid
        self.n_agents = n_agents
        self.time = initial_time
        self.end_time = end_time
        self.period_time = period_time_simulated

    async def on_start(self):
        self.logger.info("Waiting for all agents to enter the group...")
        while not len(await self.agent.group_members(self.group_jid)) >= self.n_agents:
            await _asyncio.sleep(1)
        self.current_period = 0
        self.logger.info("Starting simulation...")

    async def run(self):
        msg = Message()
        msg.to = self.group_jid
        if self.time >= self.end_time:
            msg.set_metadata("sync", "stop")
            self.kill()
        else:
            self.logger.info(f"Period {self.current_period} ({self.time})")
            msg.body = f"Period {self.current_period} ({self.time})"
            msg.set_metadata("sync", "step")
            msg.set_metadata("period", str(self.current_period))
            msg.set_metadata("time", datetime.strftime(self.time, "%Y-%m-%d %H:%M:%S"))
        await self.send_to_group(msg)
        self.current_period += 1
        self.time += self.period_time

    async def on_end(self):
        self.logger.info("Ending simulation...")
        await self.agent.stop()



"""PEAK - Python-based framework for heterogeneous agent communities.

PEAK is a framework for developing communities of heterogeneous agents.
This communities are multi-agent systems that can coexist and exchange
resources and information with each other easly.

isort: skip_file
"""

from peak.core import *
from peak.message import *
from peak.template import *
from peak.agents import *
from peak.behaviours import *
from peak.logging import *

__author__ = "GECAD"
__email__ = "brgri@isep.ipp.pt"
__version__ = "1.2.4"



import json
import logging
from typing import Callable

from peak import DF, Message, Template
from peak.core import OneShotBehaviour


class SearchCommunity(OneShotBehaviour):
    """Searches for a community."""

    def __init__(
        self, tags: list[str], callback: Callable[[list[str]], None], *args, **kargs
    ):
        super().__init__()
        self.tags = tags
        self.callback = callback
        self.args = args
        self.kargs = kargs

    async def on_start(self):
        template = Template()
        template.set_metadata("resource", "searchgroup")
        self.set_template(template)

    async def run(self):
        msg = Message()
        msg.to = DF.name(self.agent.jid.domain)
        msg.set_metadata("resource", "searchgroup")
        msg.set_metadata("tags", json.dumps(self.tags))
        await self.send(msg)
        res = await self.receive(60)
        if res is None:
            raise Exception("DF did not respond")
        communities = json.loads(res.get_metadata("communities"))
        logging.getLogger(self.__class__.__name__).debug(
            f"search: {str(self.tags)}, result: {str(communities)}"
        )
        self.callback(self.tags, communities, *self.args, **self.kargs)



import asyncio
import importlib
import logging
import os
import sys
import time
from multiprocessing import Process
from pathlib import Path
from typing import List, Type

from aioxmpp import JID, node

from peak import (
    configure_debug_mode,
    configure_multiple_agent_logging,
    configure_single_agent_logging,
)

_logger = logging.getLogger(__name__)


def bootloader(agents: list[dict]):
    if len(agents) == 1:
        boot_single_agent(agents[0])
    else:
        boot_several_agents(agents)


def boot_single_agent(agent: dict):
    _logger.info(f"booting single agent: {agent['jid']}")
    # configure_single_agent_logging()
    boot_agent(**agent, single_agent=True)


def boot_several_agents(agents: list[dict]):
    _logger.info(f"booting {len(agents)} agents (multiprocess)")
    # configure_multiple_agent_logging()
    procs: List[Process] = []
    for i, agent in enumerate(agents):
        proc = Process(
            target=boot_agent,
            kwargs=agent,
            daemon=False,
            name=agents[i]["jid"].localpart,
        )
        proc.start()
        procs.append(proc)
    _logger.info(f"all {len(agents)} processes created")
    asyncio.run(_wait_for_processes(procs))


def boot_agent(
    file: Path,
    jid: JID,
    cid: int,
    log_level: str,
    log_folder: Path,
    log_file_mode: str,
    verify_security: bool,
    debug_mode: bool,
    port: int,
    single_agent: bool = False,
    *args,
    **kargs,
):
    """Configures logging system and boots the agent.

    Args:
        file: File path where the agent's class is.
        jid: JID of the agent.
        name: The name of the agent.
        cid: Clone ID, zero if its the original.
        verify_security: If true it validates the SSL certificates.
    """
    try:
        log_file_name: str = jid.localpart + (
            f"_{jid.resource}" if jid.resource else ""
        )
        log_file = log_folder.joinpath(f"{log_file_name}.log")
        os.makedirs(log_folder, exist_ok=True)
        if single_agent:
            configure_single_agent_logging(log_level, log_file, log_file_mode)
        else:
            configure_multiple_agent_logging(log_level, log_file, log_file_mode)
        if debug_mode:
            configure_debug_mode(log_level, log_file, log_file_mode)
        _logger.info(f"instanciating agent {jid.localpart} from file {file}")
        # add port to aioxmpp
        node.discover_connectors = _change_discover_connectors_port(
            node.discover_connectors, port
        )
        agent_class = _get_class(file)
        agent_instance = agent_class(jid, cid, verify_security)
        _logger.info(f"starting agent {jid.localpart}")
        agent_instance.start().result()
        while agent_instance.is_alive():
            time.sleep(1)
        _logger.info(f"agent {jid.localpart} terminated")
    # except Exception as error:
    #    _logger.critical(f"agent {jid.localpart} terminated ({error.__class__.__name__}: {error})", exc_info=True)
    #    raise SystemExit(1)
    except KeyboardInterrupt:
        _logger.info(f"agent {jid.localpart} terminated (KeyboardInterrupt)")


def _get_class(file: Path) -> Type:
    """Gets class from a file.

    Reads a python module and retrieves the class with the name of the file.


    Args:
        file: Python module. Must have a class with the same name as the file.
            Example: agent123.py --> class agent123(...)

    Returns:
        A class object with the same name as the file.
    """
    if not file.is_file():
        raise FileNotFoundError(
            f"cannot instatiate agent from a non-existing file ({file})"
        )
    module_path, module_file = os.path.split(file.absolute())
    module_name = module_file.split(".")[0]
    sys.path.append(module_path)
    module = importlib.import_module(module_name)
    return getattr(module, module_name)


async def _wait_for_processes(processes):
    def join_process(process):
        process.join()
        if process.exitcode != 0:
            _logger.error(
                f"{process.name}'s process ended with exitcode {process.exitcode}."
            )

    await asyncio.gather(
        *[asyncio.to_thread(join_process, process) for process in processes]
    )


def _change_discover_connectors_port(original_function, new_port):
    async def new_function(*args, **kwargs):
        server_list = await original_function(*args, **kwargs)
        new_server_list = []
        for domain, _, connector in server_list:
            new_server_list.append(
                (domain, new_port, connector)  # Change the port to the new one
            )
        return new_server_list

    return new_function



import aioxmpp
import aioxmpp.forms.xso as forms_xso
from aioxmpp import MessageType
from spade.message import SPADE_X_METADATA
from spade.message import Message as _SpadeMessage

from peak.logging import getLogger

logger = getLogger(__name__)


class Message(_SpadeMessage):
    def prepare(self) -> aioxmpp.Message:
        """
        Returns an aioxmpp.stanza.Message built from the Message and prepared to be sent.

        Returns:
          aioxmpp.stanza.Message: the message prepared to be sent

        """
        msg = aioxmpp.stanza.Message(
            to=self.to,
            from_=self.sender,
            type_=MessageType.CHAT,
        )

        msg.body[None] = self.body

        # Send metadata using xep-0004: Data Forms (https://xmpp.org/extensions/xep-0004.html)
        if len(self.metadata) or self.thread:
            data = forms_xso.Data(type_=forms_xso.DataType.FORM)

            for name, value in self.metadata.items():
                data.fields.append(
                    forms_xso.Field(
                        var=name,
                        type_=forms_xso.FieldType.TEXT_SINGLE,
                        values=[value],
                    )
                )

            if self.thread:
                data.fields.append(
                    forms_xso.Field(
                        var="_thread_node",
                        type_=forms_xso.FieldType.TEXT_SINGLE,
                        values=[self.thread],
                    )
                )

            data.title = SPADE_X_METADATA
            msg.xep0004_data = [data]

        return msg



import json

from peak import CyclicBehaviour, Template, getLogger

logger = getLogger(__name__)


class CreateGraph(CyclicBehaviour):
    """Handles the requests to create graphs"""

    async def on_start(self):
        template = Template()
        template.set_metadata("resource", "graph")
        template.set_metadata("action", "create")
        self.set_template(template)

    async def run(self) -> None:
        msg = await self.receive(60)
        if msg:
            logger.debug(msg)
            id = msg.get_metadata("id")
            graph = json.loads(msg.get_metadata("graph"))
            self.agent.dataanalysis_data[id] = graph
